﻿/* -LICENSE-START-
 ** Copyright (c) 2018 Blackmagic Design
 **
 ** Permission is hereby granted, free of charge, to any person or organization
 ** obtaining a copy of the software and accompanying documentation covered by
 ** this license (the "Software") to use, reproduce, display, distribute,
 ** execute, and transmit the Software, and to prepare derivative works of the
 ** Software, and to permit third-parties to whom the Software is furnished to
 ** do so, all subject to the following:
 **
 ** The copyright notices in the Software and this entire statement, including
 ** the above license grant, this restriction and the following disclaimer,
 ** must be included in all copies of the Software, in whole or in part, and
 ** all derivative works of the Software, unless such copies or derivative
 ** works are solely in the form of machine-executable object code generated by
 ** a source language processor.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 ** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 ** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 ** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 ** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 ** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 ** DEALINGS IN THE SOFTWARE.
 ** -LICENSE-END-
 */
using System;
using DeckLinkAPI;

namespace DeckLinkCapture
{
    #region eventargs
    public class DeckLinkInputFormatChangedEventArgs : EventArgs
    {
        public readonly _BMDVideoInputFormatChangedEvents notificationEvents;
        public readonly _BMDDisplayMode displayMode;
        public readonly _BMDPixelFormat pixelFormat;

        public DeckLinkInputFormatChangedEventArgs(_BMDVideoInputFormatChangedEvents notificationEvents, _BMDDisplayMode displayMode, _BMDPixelFormat pixelFormat)
        {
            this.notificationEvents = notificationEvents;
            this.displayMode = displayMode;
            this.pixelFormat = pixelFormat;
        }
    }

    public class DeckLinkVideoFrameArrivedEventArgs : EventArgs
    {
        public readonly IDeckLinkVideoInputFrame videoFrame;
        public readonly bool inputInvalid;

        public DeckLinkVideoFrameArrivedEventArgs(IDeckLinkVideoInputFrame videoFrame, bool inputInvalid)
        {
            this.videoFrame = videoFrame;
            this.inputInvalid = inputInvalid;
        }
    }
    #endregion

    public class DeckLinkInputDevice : IDeckLinkInputCallback
    {
        private IDeckLinkInput m_deckLinkInput;
        private bool m_prevInputSignalAbsent = false;
        private int frameCount = 0;
        private long timeScale = 1001;

        public DeckLinkInputDevice(IDeckLink deckLink)
        {
            // Query input interface
            m_deckLinkInput = deckLink as IDeckLinkInput;
        }

        public event EventHandler<DeckLinkInputFormatChangedEventArgs> InputFormatChangedHandler;
        public event EventHandler<DeckLinkVideoFrameArrivedEventArgs> VideoFrameArrivedHandler;

        public void StartCapture()
        {
            var videoInputFlags = _BMDVideoInputFlags.bmdVideoInputFlagDefault | _BMDVideoInputFlags.bmdVideoInputEnableFormatDetection;

            // Set capture callback
            m_deckLinkInput.SetCallback(this);

            // Set the video input mode
            m_deckLinkInput.EnableVideoInput(_BMDDisplayMode.bmdModeNTSC, _BMDPixelFormat.bmdFormat8BitYUV, videoInputFlags);

            // Start the capture
            m_deckLinkInput.StartStreams();
        }

        void IDeckLinkInputCallback.VideoInputFormatChanged(_BMDVideoInputFormatChangedEvents notificationEvents, IDeckLinkDisplayMode newDisplayMode, _BMDDetectedVideoInputFormatFlags detectedSignalFlags)
        {
            // Restart capture with the new video mode if told to
            var pixelFormat = _BMDPixelFormat.bmdFormat8BitYUV;
            if (detectedSignalFlags.HasFlag(_BMDDetectedVideoInputFormatFlags.bmdDetectedVideoInputRGB444))
                pixelFormat = _BMDPixelFormat.bmdFormat8BitBGRA;

            // Stop the capture
            m_deckLinkInput.StopStreams();

            var displayMode = newDisplayMode.GetDisplayMode();

            // Set the video input mode
            m_deckLinkInput.EnableVideoInput(displayMode, pixelFormat, _BMDVideoInputFlags.bmdVideoInputEnableFormatDetection);

            // Start the capture
            m_deckLinkInput.StartStreams();

            string displayModeStr;
            newDisplayMode.GetName(out displayModeStr);

            long timeValue;
            newDisplayMode.GetFrameRate(out timeValue, out timeScale);

            Console.WriteLine("Video input display mode changed to " + displayModeStr);

            // Register input format changed event
            var handler = InputFormatChangedHandler;
            if (handler != null)
            {
                handler.Invoke(this, new DeckLinkInputFormatChangedEventArgs(notificationEvents, displayMode, pixelFormat)); 
            }
        }

        void IDeckLinkInputCallback.VideoInputFrameArrived(IDeckLinkVideoInputFrame videoFrame, IDeckLinkAudioInputPacket audioPacket)
        {
            if (videoFrame != null)
            {
                bool inputSignalAbsent = videoFrame.GetFlags().HasFlag(_BMDFrameFlags.bmdFrameHasNoInputSource);

                // Detect change in input signal, restart stream when valid stream detected 
                bool restartStream = !inputSignalAbsent && m_prevInputSignalAbsent;
                if (restartStream)
                {
                    m_deckLinkInput.StopStreams();
                    m_deckLinkInput.FlushStreams();
                    m_deckLinkInput.StartStreams();
                }
                m_prevInputSignalAbsent = inputSignalAbsent;

                var frameWidth = videoFrame.GetWidth();
                var frameHeight = videoFrame.GetHeight();

                long frameTime, frameDuration;
                videoFrame.GetStreamTime(out frameTime, out frameDuration, timeScale);
                Console.WriteLine(value: "Frame received #" + frameCount++ + "; Frame time = " + frameTime + 
                    "; Frame size = " + frameWidth + "x" + frameHeight + "; " + (inputSignalAbsent ? "Invalid" : "Valid")
                    + (restartStream ? " - restarting" : ""));

                // Register video frame received event
                var handler = VideoFrameArrivedHandler;
                if (handler != null)
                {
                    handler.Invoke(this, new DeckLinkVideoFrameArrivedEventArgs(videoFrame, inputSignalAbsent));
                }
            }

            System.Runtime.InteropServices.Marshal.ReleaseComObject(videoFrame);
        }

        public void StopCapture()
        {
            // Remove all listenders
            InputFormatChangedHandler = null;
            VideoFrameArrivedHandler = null;

            // Stop the capture
            m_deckLinkInput.StopStreams();

            // Disable video input
            m_deckLinkInput.DisableVideoInput();

            // Disable callbacks
            m_deckLinkInput.SetCallback(null);
        }
    }
}
